#ident "@(#) $Id: mpswrite.c,v 1.3 2002/07/03 13:02:04 bzfkocht Exp $"
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*                                                                           */
/*   File....: mpswrite.c                                                    */
/*   Name....: MPS File Write                                                */
/*   Author..: Thorsten Koch                                                 */
/*   Copyright by Author, All rights reserved                                */
/*                                                                           */
/* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
/*
 * Copyright (C) 2001 by Thorsten Koch <koch@zib.de>
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include "mshell.h"
#include "portab.h"
#include "mme.h"
#include "rbt.h"
#include "lpstore.h"

static void write_data(
   FILE*        fp,
   Bool         use_field5,
   const int    indicator1,
   const int    indicator2,
   const char*  name1,
   const char*  name2,
   const double value)
{
   assert(name1 != NULL);
   assert(name2 != NULL);
   
   if (use_field5)
   {
      fprintf(fp, " %c%c %-8.8s  %-8.8s  %12g\n",
         toupper(indicator1), toupper(indicator2), name1, name2, value);
   }
   else
   {
      fprintf(fp, " %c%c %-8.8s  %-8.8s\n",
         toupper(indicator1), toupper(indicator2), name1, name2);
   }
}   

static void write_vars(
   const Lps* lp,
   FILE*      fp,
   VarType    vartype)
{
   const Var*  var;
   const Nzo*  nzo;
   char  vtmp  [13];
   char  ctmp  [13];
   
   assert(lp != NULL);
   assert(fp != NULL);

   for(var = lp->var_root; var != NULL; var = var->next)
   {
      if (var->type != vartype)
         continue;

      if (fabs(var->cost) > EPS_ZERO)
      {
         lps_makename(vtmp, MPS_NAME_LEN + 1, var->name, var->number);

         write_data(fp, TRUE, ' ', ' ', vtmp, "OBJECTIV",
            (lp->direct == LP_MIN) ? var->cost : -var->cost);
      }
      for(nzo = var->first; nzo != NULL; nzo = nzo->var_next)
      {
         assert(nzo->var == var);
         assert(fabs(nzo->value) > EPS_ZERO);

         lps_makename(vtmp, MPS_NAME_LEN + 1, var->name, var->number);
         lps_makename(ctmp, MPS_NAME_LEN + 1,
            nzo->con->name, nzo->con->number);

         write_data(fp, TRUE, ' ', ' ', vtmp, ctmp, nzo->value);
      }
   }   
}

void mps_write(
   const Lps* lp,
   FILE*      fp)
{
   const Var*  var;
   const Con*  con;
   char  vtmp  [13];
   char  ctmp  [13];
   
   assert(lp != NULL);
   assert(fp != NULL);

   fprintf(fp, "* This file was automatically generated by zimpl\n");
   fprintf(fp, "NAME        %8.8s\n", lp->name);
   fprintf(fp, "ROWS\n");
   
   write_data(fp, FALSE, 'N', ' ', "OBJECTIV", "", 0.0);

   for(con = lp->con_root; con != NULL; con = con->next)
   {
      if (con->size > 0)
      {
         lps_makename(ctmp, MPS_NAME_LEN + 1, con->name, con->number);
         write_data(fp, FALSE, lps_getsense(con), ' ', ctmp, "", 0.0);
      }
   }
   fprintf(fp, "COLUMNS\n");

   /* First integer/binary variables
    */
   if (lp->type != LP_LP)
   {
      fprintf(fp, "    MARK0000  'MARKER'                 'INTORG'\n");

      write_vars(lp, fp, VAR_INT);
      write_vars(lp, fp, VAR_BIN);

      fprintf(fp, "    MARK0001  'MARKER'                 'INTEND'\n");
   }
   write_vars(lp, fp, VAR_CON);

   fprintf(fp, "RHS\n");
   
   for(con = lp->con_root; con != NULL; con = con->next)
   {
      if (fabs(con->rhs) > EPS_ZERO)
      {
         lps_makename(ctmp, MPS_NAME_LEN + 1, con->name, con->number);
         write_data(fp, TRUE, ' ', ' ', "RHS", ctmp, con->rhs);
      }
   }
   fprintf(fp, "BOUNDS\n");

   for(var = lp->var_root; var != NULL; var = var->next)
   {
      /* Is the variable not used at all ?
       */
      if (var->first == NULL)
         continue;
      
      /*   0, oo  -> nix
       *   l, oo  -> LO
       *   0, u   -> UP
       *   l = u  -> FX
       * -oo      -> MI
       *  oo      -> PL
       */
      lps_makename(vtmp, MPS_NAME_LEN + 1, var->name, var->number);

      if (var->lower > -INFINITY)
         write_data(fp, TRUE, 'L', 'O', "BOUND", vtmp, var->lower);
      else
         write_data(fp, FALSE, 'M', 'I', "BOUND", vtmp, 0.0);
         
      if (var->upper < INFINITY)
         write_data(fp, TRUE, 'U', 'P', "BOUND", vtmp, var->upper);
      else
         write_data(fp, FALSE, 'P', 'L', "BOUND", vtmp, 0.0);            
   }
   fprintf(fp, "ENDATA\n");
}   


